@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2013-2025 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -26,6 +26,7 @@ import codecs
 import inspect
 import re
 import signal as _signal
+import sys
 from types import MappingProxyType
 from typing import TYPE_CHECKING, Any, AnyStr, Awaitable, Callable
 from typing import Dict, Generic, Iterable, List, Mapping, Optional
@@ -45,9 +46,9 @@ from .editor import SSHLineEditorChannel, SSHLineEditorSession
 
 from .logging import SSHLogger
 
-from .misc import ChannelOpenError, MaybeAwait, ProtocolError
+from .misc import ChannelOpenError, EnvMap, MaybeAwait, ProtocolError
 from .misc import TermModes, TermSize, TermSizeArg
-from .misc import get_symbol_names, map_handler_name
+from .misc import decode_env, encode_env, get_symbol_names, map_handler_name
 
 from .packet import Boolean, Byte, String, UInt32, SSHPacket, SSHPacketHandler
 
@@ -115,7 +116,9 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
         self._send_high_water: int
         self._send_low_water: int
 
-        self._env: Dict[str, str] = {}
+        self._env: Dict[bytes, bytes] = {}
+        self._str_env: Optional[Dict[str, str]] = None
+
         self._command: Optional[str] = None
         self._subsystem: Optional[str] = None
 
@@ -143,8 +146,7 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
 
         self._recv_chan: Optional[int] = conn.add_channel(self)
 
-        self._logger = conn.logger.get_child(context='chan=%d' %
-                                             self._recv_chan)
+        self._logger = conn.logger.get_child(context=f'chan={self._recv_chan}')
 
         self.set_encoding(encoding, errors)
         self.set_write_buffer_limits()
@@ -224,12 +226,18 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
             self._request_waiters = []
 
         if self._session is not None:
-            self._session.connection_lost(exc)
+            # pylint: disable=broad-except
+            try:
+                self._session.connection_lost(exc)
+            except Exception:
+                self.logger.debug1('Uncaught exception in session ignored',
+                                   exc_info=sys.exc_info)
+
             self._session = None
 
         self._close_event.set()
 
-        if self._conn and self._recv_state == 'closed': # pragma: no branch
+        if self._conn: # pragma: no branch
             self.logger.info('Channel closed%s',
                              ': ' + str(exc) if exc else '')
 
@@ -263,8 +271,7 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
         # If recv is close_pending, we know send is already closed
         if self._recv_state == 'close_pending':
             self._recv_state = 'closed'
-
-        self._loop.call_soon(self._cleanup)
+            self._loop.call_soon(self._cleanup)
 
     async def _start_reading(self) -> None:
         """Start processing data on a new connection"""
@@ -413,11 +420,7 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
         handler = cast(_RequestHandler, getattr(self, name, None))
 
         if handler:
-            if self._session:
-                result = cast(Optional[bool], handler(packet))
-            else:
-                # Ignore requests received after application closes the channel
-                result = True
+            result = cast(Optional[bool], handler(packet))
         else:
             self.logger.debug1('Received unknown channel request: %s', request)
             result = False
@@ -880,8 +883,8 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
             low = high // 4
 
         if not 0 <= low <= high:
-            raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
-                             (high, low))
+            raise ValueError(f'high (high) must be >= low ({low}) '
+                             'must be >= 0')
 
         self.logger.debug1('Set write buffer limits: low-water=%d, '
                            'high-water=%d', low, high)
@@ -935,7 +938,7 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
         datalen = len(encoded_data)
 
         if datatype:
-            typename = ' to %s' % _data_type_names[datatype]
+            typename = f' to {_data_type_names[datatype]}'
         else:
             typename = ''
 
@@ -1035,12 +1038,41 @@ class SSHChannel(Generic[AnyStr], SSHPacketHandler):
         """Return the environment for this session
 
            This method returns the environment set by the client when
-           the session was opened. On the server, calls to this method
-           should only be made after :meth:`session_started
-           <SSHServerSession.session_started>` has been called on the
-           :class:`SSHServerSession`. When using the stream-based API,
-           calls to this can be made at any time after the handler
-           function has started up.
+           the session was opened. Keys and values are of type `str`
+           and this object only provides access to keys and values sent
+           as valid UTF-8 strings. Use :meth:`get_environment_bytes`
+           if you need to access environment variables with keys or
+           values containing binary data or non-UTF-8 encodings.
+
+           On the server, calls to this method should only be made after
+           :meth:`session_started <SSHServerSession.session_started>` has
+           been called on the :class:`SSHServerSession`. When using the
+           stream-based API, calls to this can be made at any time after
+           the handler function has started up.
+
+           :returns: A dictionary containing the environment variables
+                     set by the client
+
+        """
+
+        if self._str_env is None:
+            self._str_env = dict(decode_env(self._env))
+
+        return MappingProxyType(self._str_env)
+
+    def get_environment_bytes(self) -> Mapping[bytes, bytes]:
+        """Return the environment for this session
+
+           This method returns the environment set by the client when
+           the session was opened. Keys and values are of type `bytes`
+           and can include arbitrary binary data, with the exception
+           of NUL (\0) bytes.
+
+           On the server, calls to this method should only be made after
+           :meth:`session_started <SSHServerSession.session_started>` has
+           been called on the :class:`SSHServerSession`. When using the
+           stream-based API, calls to this can be made at any time after
+           the handler function has started up.
 
            :returns: A dictionary containing the environment variables
                      set by the client
@@ -1093,16 +1125,18 @@ class SSHClientChannel(SSHChannel, Generic[AnyStr]):
     _read_datatypes = {EXTENDED_DATA_STDERR}
 
     def __init__(self, conn: 'SSHClientConnection',
-                 loop: asyncio.AbstractEventLoop, encoding: Optional[str],
-                 errors: str, window: int, max_pktsize: int):
+                 loop: asyncio.AbstractEventLoop, utf8_decode_errors: str,
+                 encoding: Optional[str], errors: str, window: int,
+                 max_pktsize: int):
         super().__init__(conn, loop, encoding, errors, window, max_pktsize)
 
+        self._utf8_decode_errors = utf8_decode_errors
         self._exit_status: Optional[int] = None
         self._exit_signal: Optional[_ExitSignal] = None
 
     async def create(self, session_factory: SSHClientSessionFactory[AnyStr],
                      command: Optional[str], subsystem: Optional[str],
-                     env: Dict[str, str], request_pty: bool,
+                     env: Dict[bytes, bytes], request_pty: bool,
                      term_type: Optional[str], term_size: TermSizeArg,
                      term_modes: TermModes, x11_forwarding: Union[bool, str],
                      x11_display: Optional[str], x11_auth_path: Optional[str],
@@ -1124,10 +1158,16 @@ class SSHClientChannel(SSHChannel, Generic[AnyStr]):
         self._command = command
         self._subsystem = subsystem
 
-        for name, env_value in env.items():
-            self.logger.debug1('  Env: %s=%s', name, env_value)
-            self._send_request(b'env', String(str(name)),
-                               String(str(env_value)))
+        for key, value in env.items():
+            self.logger.debug1('  Env: %s=%s', key, value)
+
+            if not isinstance(key, (bytes, str)):
+                key = str(key)
+
+            if not isinstance(value, (bytes, str)):
+                value = str(value)
+
+            self._send_request(b'env', String(key), String(value))
 
         if request_pty:
             self.logger.debug1('  Terminal type: %s', term_type or 'None')
@@ -1149,7 +1189,7 @@ class SSHClientChannel(SSHChannel, Generic[AnyStr]):
             modes = b''
             for mode, mode_value in term_modes.items():
                 if mode <= PTY_OP_END or mode >= PTY_OP_RESERVED:
-                    raise ValueError('Invalid pty mode: %s' % mode)
+                    raise ValueError(f'Invalid pty mode: {mode}')
 
                 name = _pty_mode_names.get(mode, str(mode))
                 self.logger.debug2('  Mode %s: %d', name, mode_value)
@@ -1261,7 +1301,7 @@ class SSHClientChannel(SSHChannel, Generic[AnyStr]):
 
         try:
             signal = signal_bytes.decode('ascii')
-            msg = msg_bytes.decode('utf-8')
+            msg = msg_bytes.decode('utf-8', self._utf8_decode_errors)
             lang = lang_bytes.decode('ascii')
         except UnicodeDecodeError:
             raise ProtocolError('Invalid exit signal request') from None
@@ -1407,7 +1447,7 @@ class SSHClientChannel(SSHChannel, Generic[AnyStr]):
             try:
                 signal = _signal_names[signal]
             except KeyError:
-                raise ValueError('Unknown signal: %s' % int(signal)) from None
+                raise ValueError(f'Unknown signal: {signal}') from None
 
         self.logger.info('Sending %s signal', signal)
 
@@ -1465,8 +1505,8 @@ class SSHServerChannel(SSHChannel, Generic[AnyStr]):
 
         super().__init__(conn, loop, encoding, errors, window, max_pktsize)
 
-        self._env = cast(Dict[str, str],
-                         conn.get_key_option('environment', {}))
+        env_opt = cast(EnvMap, conn.get_key_option('environment', {}))
+        self._env = dict(encode_env(env_opt))
 
         self._allow_pty = allow_pty
         self._line_editor = line_editor
@@ -1621,19 +1661,12 @@ class SSHServerChannel(SSHChannel, Generic[AnyStr]):
     def _process_env_request(self, packet: SSHPacket) -> bool:
         """Process a request to set an environment variable"""
 
-        name_bytes = packet.get_string()
-        value_bytes = packet.get_string()
+        key = packet.get_string()
+        value = packet.get_string()
         packet.check_end()
 
-        try:
-            name = name_bytes.decode('utf-8')
-            value = value_bytes.decode('utf-8')
-        except UnicodeDecodeError:
-            self.logger.debug1('Invalid environment data')
-            return False
-
-        self.logger.debug1('  Env: %s=%s', name, value)
-        self._env[name] = value
+        self.logger.debug1('  Env: %s=%s', key, value)
+        self._env[key] = value
         return True
 
     def _start_session(self, command: Optional[str] = None,
@@ -2034,16 +2067,16 @@ class SSHTCPChannel(SSHForwardChannel, Generic[AnyStr]):
             SSHTCPSession[AnyStr]:
         """Create a new outbound TCP session"""
 
-        return (await self._open_tcp(session_factory, b'direct-tcpip',
-                                     host, port, orig_host, orig_port))
+        return await self._open_tcp(session_factory, b'direct-tcpip',
+                                    host, port, orig_host, orig_port)
 
     async def accept(self, session_factory: SSHTCPSessionFactory[AnyStr],
                      host: str, port: int, orig_host: str,
                      orig_port: int) -> SSHTCPSession[AnyStr]:
         """Create a new forwarded TCP session"""
 
-        return (await self._open_tcp(session_factory, b'forwarded-tcpip',
-                                     host, port, orig_host, orig_port))
+        return await self._open_tcp(session_factory, b'forwarded-tcpip',
+                                    host, port, orig_host, orig_port)
 
     def set_inbound_peer_names(self, dest_host: str, dest_port: int,
                                orig_host: str, orig_port: int) -> None:
