@@ -1,4 +1,4 @@
-# Copyright (c) 2020-2021 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2020-2024 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -24,15 +24,34 @@ import os
 import re
 import shlex
 import socket
+import subprocess
 
 from hashlib import sha1
 from pathlib import Path, PurePath
+from subprocess import DEVNULL
+from typing import Callable, Dict, List, NoReturn, Optional, Sequence
+from typing import Set, Tuple, Union, cast
 
 from .constants import DEFAULT_PORT
-from .misc import ip_address
+from .logging import logger
+from .misc import DefTuple, FilePath, ip_address
 from .pattern import HostPatternList, WildcardPatternList
 
 
+ConfigPaths = Union[None, FilePath, Sequence[FilePath]]
+
+
+_token_pattern = re.compile(r'%(.)')
+_env_pattern = re.compile(r'\${(.*)}')
+
+
+def _exec(cmd: str) -> bool:
+    """Execute a command and return if exit status is 0"""
+
+    return subprocess.run(cmd, check=False, shell=True, stdin=DEVNULL,
+                          stdout=DEVNULL, stderr=DEVNULL).returncode == 0
+
+
 class ConfigParseError(ValueError):
     """Configuration parsing exception"""
 
@@ -41,75 +60,83 @@ class SSHConfig:
     """Settings from an OpenSSH config file"""
 
     _conditionals = {'match'}
-    _no_split = set()
+    _no_split: Set[str] = set()
     _percent_expand = {'AuthorizedKeysFile'}
-    _handlers = {}
+    _handlers: Dict[str, Tuple[str, Callable]] = {}
 
-    def __init__(self, last_config, reload, *args, **kwargs):
+    def __init__(self, last_config: Optional['SSHConfig'], reload: bool,
+                 canonical: bool, final: bool):
         if last_config:
             self._last_options = last_config.get_options(reload)
         else:
             self._last_options = {}
 
-        self._path = ''
+        self._canonical = canonical
+        self._final = True if final else None
+        self._default_path = Path('~', '.ssh').expanduser()
+        self._path = Path()
         self._line_no = 0
         self._matching = True
         self._options = self._last_options.copy()
-        self._tokens = {}
+        self._tokens: Dict[str, str] = {}
 
-        self._init_options(*args, **kwargs)
+        self.loaded = False
 
-    def _init_options(self):
-        """Set options used for matching and token substitutions"""
-
-        raise NotImplementedError
-
-    def _error(self, reason, *args):
+    def _error(self, reason: str) -> NoReturn:
         """Raise a configuration parsing error"""
 
-        raise ConfigParseError('%s line %s: %s' % (self._path, self._line_no,
-                                                   reason % args))
+        raise ConfigParseError(f'{self._path} line {self._line_no}: {reason}')
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         raise NotImplementedError
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         raise NotImplementedError
 
-    def _expand_val(self, value):
-        """Perform percent token expansion on a string"""
+    def _expand_token(self, match):
+        """Expand a percent token reference"""
 
-        last_idx = 0
-        result = []
+        try:
+            token = match.group(1)
+            return self._tokens[token]
+        except KeyError:
+            if token == 'd':
+                raise ConfigParseError('Home directory is '
+                                       'not available') from None
+            elif token == 'i':
+                raise ConfigParseError('User id not available') from None
+            else:
+                raise ConfigParseError('Invalid token expansion: ' +
+                                       token) from None
 
-        for match in re.finditer(r'%', value):
-            idx = match.start()
+    @staticmethod
+    def _expand_env(match):
+        """Expand an environment variable reference"""
 
-            if idx < last_idx:
-                continue
+        try:
+            var = match.group(1)
+            return os.environ[var]
+        except KeyError:
+            raise ConfigParseError('Invalid environment expansion: ' +
+                                   var) from None
 
-            try:
-                result.extend([value[last_idx:idx],
-                               self._tokens[value[idx+1]]])
-                last_idx = idx + 2
-            except IndexError:
-                raise ConfigParseError('Invalid token substitution') from None
-            except KeyError:
-                raise ConfigParseError('Invalid token substitution: %s' %
-                                       value[idx+1]) from None
+    def _expand_val(self, value: str) -> str:
+        """Perform percent token and environment expansion on a string"""
 
-        result.append(value[last_idx:])
-        return ''.join(result)
+        return _env_pattern.sub(self._expand_env,
+                                _token_pattern.sub(self._expand_token, value))
 
-    def _include(self, option, args):
+    def _include(self, option: str, args: List[str]) -> None:
         """Read config from a list of other config files"""
 
         # pylint: disable=unused-argument
 
+        old_path = self._path
+
         for pattern in args:
             path = Path(pattern).expanduser()
 
@@ -117,143 +144,208 @@ class SSHConfig:
                 pattern = str(Path(*path.parts[1:]))
                 path = Path(path.anchor)
             else:
-                path = Path(self._path).parent
+                path = self._default_path
 
-            for path in path.glob(pattern):
+            paths = list(p for p in path.glob(pattern) if p.is_file())
+
+            if not paths:
+                logger.debug1(f'Config pattern "{pattern}" matched no files')
+
+            for path in paths:
                 self.parse(path)
 
+        self._path = old_path
         args.clear()
 
-    def _match(self, option, args):
+    def _match(self, option: str, args: List[str]) -> None:
         """Begin a conditional block"""
 
         # pylint: disable=unused-argument
 
+        matching = True
+
         while args:
             match = args.pop(0).lower()
 
+            if match[0] == '!':
+                match = match[1:]
+                negated = True
+            else:
+                negated = False
+
+            if match == 'final' and self._final is None:
+                self._final = False
+
             if match == 'all':
-                self._matching = True
-                continue
+                result = True
+            elif match == 'canonical':
+                result = self._canonical
+            elif match == 'final':
+                result = cast(bool, self._final)
+            else:
+                match_val = self._match_val(match)
 
-            match_val = self._match_val(match)
+                if match != 'exec' and match_val is None:
+                    self._error(f'Invalid match condition {match}')
 
-            if match_val is None:
-                self._error('Invalid match condition')
+                try:
+                    arg = args.pop(0)
+                except IndexError:
+                    self._error(f'Missing {match} match pattern')
 
-            try:
-                if match in ('address', 'localaddress'):
-                    pattern = HostPatternList(args.pop(0))
-                    ip = ip_address(match_val) if match_val else None
-                    self._matching = pattern.matches(None, match_val, ip)
-                else:
-                    pattern = WildcardPatternList(args.pop(0))
-                    self._matching = pattern.matches(match_val)
-            except IndexError:
-                self._error('Missing %s match pattern', match)
+                if matching:
+                    if match == 'exec':
+                        result = _exec(arg)
+                    elif match in ('address', 'localaddress'):
+                        host_pat = HostPatternList(arg)
+                        ip = ip_address(cast(str, match_val)) \
+                            if match_val else None
+                        result = host_pat.matches(None, match_val, ip)
+                    else:
+                        wild_pat = WildcardPatternList(arg)
+                        result = wild_pat.matches(match_val)
 
-            if not self._matching:
-                args.clear()
-                break
+            if matching and result == negated:
+                matching = False
 
-    def _set_bool(self, option, args):
+        self._matching = matching
+
+    def _set_bool(self, option: str, args: List[str]) -> None:
         """Set a boolean config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value in ('yes', 'true'):
+        if value_str in ('yes', 'true'):
             value = True
-        elif value in ('no', 'false'):
+        elif value_str in ('no', 'false'):
             value = False
         else:
-            self._error('Invalid %s boolean value: %s', option, value)
+            self._error(f'Invalid {option} boolean value: {value_str}')
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_int(self, option, args):
+    def _set_bool_or_str(self, option: str, args: List[str]) -> None:
+        """Set a boolean or string config option"""
+
+        value_str = args.pop(0)
+        value_lower = value_str.lower()
+
+        if value_lower in ('yes', 'true'):
+            value: Union[bool, str] = True
+        elif value_lower in ('no', 'false'):
+            value = False
+        else:
+            value = value_str
+
+        if option not in self._options:
+            self._options[option] = value
+
+    def _set_int(self, option: str, args: List[str]) -> None:
         """Set an integer config option"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
         try:
-            value = int(value)
+            value = int(value_str)
         except ValueError:
-            self._error('Invalid %s integer value: %s', option, value)
+            self._error(f'Invalid {option} integer value: {value_str}')
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_string(self, option, args):
+    def _set_string(self, option: str, args: List[str]) -> None:
         """Set a string config option"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
-        if value.lower() == 'none':
+        if value_str.lower() == 'none':
             value = None
+        else:
+            value = value_str
 
         if option not in self._options:
             self._options[option] = value
 
-    def _append_string(self, option, args):
+    def _append_string(self, option: str, args: List[str]) -> None:
         """Append a string config option to a list"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
-        if value.lower() != 'none':
+        if value_str.lower() != 'none':
             if option in self._options:
-                self._options[option].append(value)
+                cast(List[str], self._options[option]).append(value_str)
             else:
-                self._options[option] = [value]
+                self._options[option] = [value_str]
         else:
             if option not in self._options:
                 self._options[option] = []
 
-    def _set_string_list(self, option, args):
+    def _set_string_list(self, option: str, args: List[str]) -> None:
         """Set whitespace-separated string config options as a list"""
 
         if option not in self._options:
-            self._options[option] = args[:]
+            if len(args) == 1 and args[0].lower() == 'none':
+                self._options[option] = []
+            else:
+                self._options[option] = args[:]
 
         args.clear()
 
-    def _append_string_list(self, option, args):
+    def _append_string_list(self, option: str, args: List[str]) -> None:
         """Append whitespace-separated string config options to a list"""
 
         if option in self._options:
-            self._options[option].extend(args)
+            cast(List[str], self._options[option]).extend(args)
         else:
             self._options[option] = args[:]
 
         args.clear()
 
-    def _set_address_family(self, option, args):
+    def _set_address_family(self, option: str, args: List[str]) -> None:
         """Set an address family config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value == 'any':
+        if value_str == 'any':
             value = socket.AF_UNSPEC
-        elif value == 'inet':
+        elif value_str == 'inet':
             value = socket.AF_INET
-        elif value == 'inet6':
+        elif value_str == 'inet6':
             value = socket.AF_INET6
         else:
-            self._error('Invalid %s value: %s', option, value)
+            self._error(f'Invalid {option} value: {value_str}')
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_rekey_limits(self, option, args):
+    def _set_canonicalize_host(self, option: str, args: List[str]) -> None:
+        """Set a canonicalize host config option"""
+
+        value_str = args.pop(0).lower()
+
+        if value_str in ('yes', 'true'):
+            value: Union[bool, str] = True
+        elif value_str in ('no', 'false'):
+            value = False
+        elif value_str == 'always':
+            value = value_str
+        else:
+            self._error(f'Invalid {option} value: {value_str}')
+
+        if option not in self._options:
+            self._options[option] = value
+
+    def _set_rekey_limits(self, option: str, args: List[str]) -> None:
         """Set rekey limits config option"""
 
-        byte_limit = args.pop(0).lower()
+        byte_limit: Union[str, Tuple[()]] = args.pop(0).lower()
 
         if byte_limit == 'default':
             byte_limit = ()
 
         if args:
-            time_limit = args.pop(0).lower()
+            time_limit: Optional[Union[str, Tuple[()]]] = args.pop(0).lower()
 
             if time_limit == 'none':
                 time_limit = None
@@ -263,7 +355,12 @@ class SSHConfig:
         if option not in self._options:
             self._options[option] = byte_limit, time_limit
 
-    def parse(self, path):
+    def has_match_final(self) -> bool:
+        """Return whether this config includes a 'Match final' block"""
+
+        return self._final is not None
+
+    def parse(self, path: Path) -> None:
         """Parse an OpenSSH config file and return matching declarations"""
 
         self._path = path
@@ -271,6 +368,8 @@ class SSHConfig:
         self._matching = True
         self._tokens = {'%': '%'}
 
+        logger.debug1('Reading config from "%s"', path)
+
         with open(path) as file:
             for line in file:
                 self._line_no += 1
@@ -280,23 +379,33 @@ class SSHConfig:
                     continue
 
                 try:
-                    args = shlex.split(line)
+                    split_args = shlex.split(line)
                 except ValueError as exc:
                     self._error(str(exc))
 
-                option = args.pop(0)
+                args = []
+                loption = ''
+                allow_equal = True
 
-                if option.endswith('='):
-                    option = option[:-1]
-                elif '=' in option:
-                    option, arg = option.split('=', 1)
-                    args[:0] =[arg]
-                elif args and args[0] == '=':
-                    del args[0]
-                elif args and args[0].startswith('='):
-                    args[0] = args[0][1:]
+                for i, arg in enumerate(split_args, 1):
+                    if arg.startswith('='):
+                        if len(arg) > 1:
+                            args.append(arg[1:])
+                    elif not allow_equal:
+                        args.extend(split_args[i-1:])
+                        break
+                    elif arg.endswith('='):
+                        args.append(arg[:-1])
+                    elif '=' in arg:
+                        arg, val = arg.split('=', 1)
+                        args.append(arg)
+                        args.append(val)
+                    else:
+                        args.append(arg)
 
-                loption = option.lower()
+                    if i == 1:
+                        loption = args.pop(0).lower()
+                        allow_equal = loption in self._conditionals
 
                 if loption in self._no_split:
                     args = [line.lstrip()[len(loption):].strip()]
@@ -310,12 +419,12 @@ class SSHConfig:
                     continue
 
                 if not args:
-                    self._error('Missing %s value', option)
+                    self._error(f'Missing {option} value')
 
                 handler(self, option, args)
 
                 if args:
-                    self._error('Extra data at end: %s', ' '.join(args))
+                    self._error(f'Extra data at end: {" ".join(args)}')
 
         self._set_tokens()
 
@@ -332,38 +441,44 @@ class SSHConfig:
 
                 self._options[option] = value
 
-    def get_options(self, reload):
+    def get_options(self, reload: bool) -> Dict[str, object]:
         """Return options to base a new config object on"""
 
         return self._last_options.copy() if reload else self._options.copy()
 
     @classmethod
-    def load(cls, last_config, config_paths, reload, *args, **kwargs):
+    def load(cls, last_config: Optional['SSHConfig'],
+             config_paths: ConfigPaths, reload: bool,
+             canonical: bool, final: bool, *args: object) -> 'SSHConfig':
         """Load a list of OpenSSH config files into a config object"""
 
-        config = cls(last_config, reload, *args, **kwargs)
+        config = cls(last_config, reload, canonical, final, *args)
 
         if config_paths:
-            if isinstance(config_paths, (str, bytes, PurePath)):
-                config_paths = [config_paths]
+            if isinstance(config_paths, (str, PurePath)):
+                paths: Sequence[FilePath] = [config_paths]
+            else:
+                paths = config_paths
 
-            for path in config_paths:
-                config.parse(path)
+            for path in paths:
+                config.parse(Path(path))
+
+            config.loaded = True
 
         return config
 
-    def get(self, option, default=None):
+    def get(self, option: str, default: object = None) -> object:
         """Get the value of a config option"""
 
         return self._options.get(option, default)
 
-    def get_compression_algs(self, default=None):
+    def get_compression_algs(self) -> DefTuple[str]:
         """Return the compression algorithms to use"""
 
         compression = self.get('Compression')
 
         if compression is None:
-            return default
+            return ()
         elif compression:
             return 'zlib@openssh.com,zlib,none'
         else:
@@ -374,14 +489,14 @@ class SSHClientConfig(SSHConfig):
     """Settings from an OpenSSH client config file"""
 
     _conditionals = {'host', 'match'}
-    _no_split = {'remotecommand'}
-    _percent_expand = {'CertificateFile', 'IdentityAgent',
+    _no_split = {'proxycommand', 'remotecommand'}
+    _percent_expand = {'CertificateFile', 'ForwardAgent', 'IdentityAgent',
                        'IdentityFile', 'ProxyCommand', 'RemoteCommand'}
 
-    def _init_options(self, local_user, user, host, port):
-        """Set options used for matching and token substitutions"""
-
-        # pylint: disable=arguments-differ
+    def __init__(self, last_config: 'SSHConfig', reload: bool,
+                 canonical: bool, final: bool, local_user: str,
+                 user: str, host: str, port: int) -> None:
+        super().__init__(last_config, reload, canonical, final)
 
         self._local_user = local_user
         self._orig_host = host
@@ -392,7 +507,7 @@ class SSHClientConfig(SSHConfig):
         if port != ():
             self._options['Port'] = port
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         if match == 'host':
@@ -403,10 +518,12 @@ class SSHClientConfig(SSHConfig):
             return self._local_user
         elif match == 'user':
             return self._options.get('User', self._local_user)
+        elif match == 'tagged':
+            return self._options.get('Tag', '')
         else:
             return None
 
-    def _match_host(self, option, args):
+    def _match_host(self, option: str, args: List[str]) -> None:
         """Begin a conditional block matching on host"""
 
         # pylint: disable=unused-argument
@@ -415,31 +532,34 @@ class SSHClientConfig(SSHConfig):
         self._matching = WildcardPatternList(pattern).matches(self._orig_host)
         args.clear()
 
-    def _set_hostname(self, option, args):
+    def _set_hostname(self, option: str, args: List[str]) -> None:
         """Set hostname config option"""
 
         value = args.pop(0)
 
         if option not in self._options:
-            self._tokens['h'] = self._options.get(option, self._orig_host)
+            self._tokens['h'] = \
+                cast(str, self._options.get(option, self._orig_host))
             self._options[option] = self._expand_val(value)
 
-    def _set_request_tty(self, option, args):
+    def _set_request_tty(self, option: str, args: List[str]) -> None:
         """Set a pseudo-terminal request config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value in ('yes', 'true'):
-            value = True
-        elif value in ('no', 'false'):
+        if value_str in ('yes', 'true'):
+            value: Union[bool, str] = True
+        elif value_str in ('no', 'false'):
             value = False
-        elif value not in ('force', 'auto'):
-            self._error('Invalid %s value: %s', option, value)
+        elif value_str in ('force', 'auto'):
+            value = value_str
+        else:
+            self._error(f'Invalid {option} value: {value_str}')
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         local_host = socket.gethostname()
@@ -447,15 +567,15 @@ class SSHClientConfig(SSHConfig):
         idx = local_host.find('.')
         short_local_host = local_host if idx < 0 else local_host[:idx]
 
-        host = self._options.get('Hostname', self._orig_host)
+        host = cast(str, self._options.get('Hostname', self._orig_host))
         port = str(self._options.get('Port', DEFAULT_PORT))
-        user = self._options.get('User') or self._local_user
+        user = cast(str, self._options.get('User') or self._local_user)
+        home = os.path.expanduser('~')
 
         conn_info = ''.join((local_host, host, port, user))
         conn_hash = sha1(conn_info.encode('utf-8')).hexdigest()
 
         self._tokens.update({'C': conn_hash,
-                             'd': str(Path.home()),
                              'h': host,
                              'L': short_local_host,
                              'l': local_host,
@@ -464,7 +584,10 @@ class SSHClientConfig(SSHConfig):
                              'r': user,
                              'u': self._local_user})
 
-        if hasattr(os, 'getuid'): # pragma: no branch
+        if home != '~':
+            self._tokens['d'] = home
+
+        if hasattr(os, 'getuid'):
             self._tokens['i'] = str(os.getuid())
 
     _handlers = {option.lower(): (option, handler) for option, handler in (
@@ -474,6 +597,11 @@ class SSHClientConfig(SSHConfig):
 
         ('AddressFamily',                   SSHConfig._set_address_family),
         ('BindAddress',                     SSHConfig._set_string),
+        ('CanonicalDomains',                SSHConfig._set_string_list),
+        ('CanonicalizeFallbackLocal',       SSHConfig._set_bool),
+        ('CanonicalizeHostname',            SSHConfig._set_canonicalize_host),
+        ('CanonicalizeMaxDots',             SSHConfig._set_int),
+        ('CanonicalizePermittedCNAMEs',     SSHConfig._set_string_list),
         ('CASignatureAlgorithms',           SSHConfig._set_string),
         ('CertificateFile',                 SSHConfig._append_string),
         ('ChallengeResponseAuthentication', SSHConfig._set_bool),
@@ -481,7 +609,7 @@ class SSHClientConfig(SSHConfig):
         ('Compression',                     SSHConfig._set_bool),
         ('ConnectTimeout',                  SSHConfig._set_int),
         ('EnableSSHKeySign',                SSHConfig._set_bool),
-        ('ForwardAgent',                    SSHConfig._set_bool),
+        ('ForwardAgent',                    SSHConfig._set_bool_or_str),
         ('ForwardX11Trusted',               SSHConfig._set_bool),
         ('GlobalKnownHostsFile',            SSHConfig._set_string_list),
         ('GSSAPIAuthentication',            SSHConfig._set_bool),
@@ -501,7 +629,7 @@ class SSHClientConfig(SSHConfig):
         ('PKCS11Provider',                  SSHConfig._set_string),
         ('PreferredAuthentications',        SSHConfig._set_string),
         ('Port',                            SSHConfig._set_int),
-        ('ProxyCommand',                    SSHConfig._set_string_list),
+        ('ProxyCommand',                    SSHConfig._set_string),
         ('ProxyJump',                       SSHConfig._set_string),
         ('PubkeyAuthentication',            SSHConfig._set_bool),
         ('RekeyLimit',                      SSHConfig._set_rekey_limits),
@@ -510,7 +638,8 @@ class SSHClientConfig(SSHConfig):
         ('SendEnv',                         SSHConfig._append_string_list),
         ('ServerAliveCountMax',             SSHConfig._set_int),
         ('ServerAliveInterval',             SSHConfig._set_int),
-        ('SetEnv',                          SSHConfig._append_string_list),
+        ('SetEnv',                          SSHConfig._set_string_list),
+        ('Tag',                             SSHConfig._set_string),
         ('TCPKeepAlive',                    SSHConfig._set_bool),
         ('User',                            SSHConfig._set_string),
         ('UserKnownHostsFile',              SSHConfig._set_string_list)
@@ -520,10 +649,10 @@ class SSHClientConfig(SSHConfig):
 class SSHServerConfig(SSHConfig):
     """Settings from an OpenSSH server config file"""
 
-    def _init_options(self, local_addr, local_port, user, host, addr):
-        """Set options used for matching and token substitutions"""
-
-        # pylint: disable=arguments-differ
+    def __init__(self, last_config: 'SSHConfig', reload: bool,
+                 canonical: bool, final: bool, local_addr: str,
+                 local_port: int, user: str, host: str, addr: str) -> None:
+        super().__init__(last_config, reload, canonical, final)
 
         self._local_addr = local_addr
         self._local_port = local_port
@@ -531,7 +660,7 @@ class SSHServerConfig(SSHConfig):
         self._host = host or addr
         self._addr = addr
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         if match == 'localaddress':
@@ -547,7 +676,7 @@ class SSHServerConfig(SSHConfig):
         else:
             return None
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         self._tokens.update({'u': self._user})
@@ -560,6 +689,11 @@ class SSHServerConfig(SSHConfig):
         ('AuthorizedKeysFile',              SSHConfig._set_string_list),
         ('AllowAgentForwarding',            SSHConfig._set_bool),
         ('BindAddress',                     SSHConfig._set_string),
+        ('CanonicalDomains',                SSHConfig._set_string_list),
+        ('CanonicalizeFallbackLocal',       SSHConfig._set_bool),
+        ('CanonicalizeHostname',            SSHConfig._set_canonicalize_host),
+        ('CanonicalizeMaxDots',             SSHConfig._set_int),
+        ('CanonicalizePermittedCNAMEs',     SSHConfig._set_string_list),
         ('CASignatureAlgorithms',           SSHConfig._set_string),
         ('ChallengeResponseAuthentication', SSHConfig._set_bool),
         ('Ciphers',                         SSHConfig._set_string),
